pancakeswap_liquidity_pool_abi=[
    {
        "inputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "Approval",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "sender",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0",
                "type": "uint256"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1",
                "type": "uint256"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            }
        ],
        "name": "Burn",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "sender",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0",
                "type": "uint256"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1",
                "type": "uint256"
            }
        ],
        "name": "Mint",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "sender",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0In",
                "type": "uint256"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1In",
                "type": "uint256"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0Out",
                "type": "uint256"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1Out",
                "type": "uint256"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            }
        ],
        "name": "Swap",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint112",
                "name": "reserve0",
                "type": "uint112"
            },
            {
                "indexed": False,
                "internalType": "uint112",
                "name": "reserve1",
                "type": "uint112"
            }
        ],
        "name": "Sync",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "Transfer",
        "type": "event"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "DOMAIN_SEPARATOR",
        "outputs": [
            {
                "internalType": "bytes32",
                "name": "",
                "type": "bytes32"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MINIMUM_LIQUIDITY",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "PERMIT_TYPEHASH",
        "outputs": [
            {
                "internalType": "bytes32",
                "name": "",
                "type": "bytes32"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "allowance",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "approve",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "balanceOf",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            }
        ],
        "name": "burn",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "amount0",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "amount1",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [
            {
                "internalType": "uint8",
                "name": "",
                "type": "uint8"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "factory",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getReserves",
        "outputs": [
            {
                "internalType": "uint112",
                "name": "_reserve0",
                "type": "uint112"
            },
            {
                "internalType": "uint112",
                "name": "_reserve1",
                "type": "uint112"
            },
            {
                "internalType": "uint32",
                "name": "_blockTimestampLast",
                "type": "uint32"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "_token0",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "_token1",
                "type": "address"
            }
        ],
        "name": "initialize",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "kLast",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            }
        ],
        "name": "mint",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "liquidity",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "nonces",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "deadline",
                "type": "uint256"
            },
            {
                "internalType": "uint8",
                "name": "v",
                "type": "uint8"
            },
            {
                "internalType": "bytes32",
                "name": "r",
                "type": "bytes32"
            },
            {
                "internalType": "bytes32",
                "name": "s",
                "type": "bytes32"
            }
        ],
        "name": "permit",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "price0CumulativeLast",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "price1CumulativeLast",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            }
        ],
        "name": "skim",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "uint256",
                "name": "amount0Out",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "amount1Out",
                "type": "uint256"
            },
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "bytes",
                "name": "data",
                "type": "bytes"
            }
        ],
        "name": "swap",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [],
        "name": "sync",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "token0",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "token1",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "transfer",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "transferFrom",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# Pip
from kw3.constants import Constants

# Local
from ..core.reserves import Reserves

# -------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------ class: ReservesWbnb ----------------------------------------------------- #

class ReservesWbnb(Reserves):

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,
        reserve0: int,
        reserve1: int,
        blockTimestampLast: int,
        is_base_wbnb: bool
    ):
        super().__init__(
            reserve0=reserve0,
            reserve1=reserve1,
            blockTimestampLast=blockTimestampLast
        )

        self.__is_base_wbnb = is_base_wbnb


    # --------------------------------------------------- Public properties -------------------------------------------------- #

    @property
    def reserveToken(self) -> int:
        return self.reserve1 if self.__is_base_wbnb else self.reserve0

    @property
    def reserveWbnb(self) -> int:
        return self.reserve0 if self.__is_base_wbnb else self.reserve1

    @property
    def token_wbnb_rate(self) -> float:
        return self.token1_token0_rate if self.__is_base_wbnb else self.token0_token1_rate

    @property
    def wbnb_token_rate(self) -> float:
        return self.token0_token1_rate if self.__is_base_wbnb else self.token1_token0_rate


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    def tokenPrice(
        self,
        token_decimals: int
    ) -> float:
        '''BNB for 1 Token'''
        token_0_decimals, token_1_decimals = (Constants.WBNB.DECIMALS, token_decimals) if self.__is_base_wbnb else (token_decimals, Constants.WBNB.DECIMALS)

        return (self.token1Price if self.__is_base_wbnb else self.token0Price)(
            token_0_decimals=token_0_decimals,
            token_1_decimals=token_1_decimals
        )

    def wbnbPrice(
        self,
        token_decimals: int
    ) -> float:
        '''Token for 1 BNB'''
        token_0_decimals, token_1_decimals = (token_decimals, Constants.WBNB.DECIMALS) if self.__is_base_wbnb else (Constants.WBNB.DECIMALS, token_decimals)

        return (self.token0Price if self.__is_base_wbnb else self.token1Price)(
            token_0_decimals=token_0_decimals,
            token_1_decimals=token_1_decimals
        )


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional
import random

# Pip
from kw3 import Web3, Bep20, Web3Wrapper as CoreWeb3Wrapper

from .pancakeswap import *

# -------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------ class: Web3Wrapper ------------------------------------------------------ #

class Web3Wrapper(CoreWeb3Wrapper):

    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    # Factory

    def pancakeswap_factory(self) -> Bep20:
        return PancakeswapFactory(self.__web3)

    def pancakeswap_wbnb_factory(self) -> PancakeswapWbnbFactory:
        return PancakeswapWbnbFactory(self.__web3)

    def pancakeswap_busd_factory(self) -> PancakeswapBusdFactory:
        return PancakeswapBusdFactory(self.__web3)


    # Liquidity pool

    def pancakeswap_liquidity_pool(
        self,
        address: str
    ) -> PancakeswapLiquidityPool:
        return PancakeswapLiquidityPool(self.__web3, address=address)

    def pancakeswap_wbnb_liquidity_pool(
        self,
        liquidity_pool_address: str
    ) -> PancakeswapWbnbLiquidityPool:
        return PancakeswapWbnbLiquidityPool(
            web3=self.__web3,
            address=liquidity_pool_address
        )

    def pancakeswap_busd_liquidity_pool(
        self,
        liquidity_pool_address: str
    ) -> PancakeswapBusdLiquidityPool:
        return PancakeswapBusdLiquidityPool(
            web3=self.__web3,
            address=liquidity_pool_address
        )

    def pancakeswap_wbnb_liquidity_pool_for_token(
        self,
        token_address: str
    ) -> PancakeswapLiquidityPool:
        return self.pancakeswap_wbnb_factory().getPair(
            address=token_address
        )

    def pancakeswap_busd_liquidity_pool_for_token(
        self,
        token_address: str
    ) -> PancakeswapLiquidityPool:
        return self.pancakeswap_busd_factory().getPair(
            address=token_address
        )

    def pancakeswap_wbnb_busd_liquidity_pool(self) -> PancakeswapWbnbBusdLiquidityPool:
        return PancakeswapWbnbBusdLiquidityPool(self.__web3)


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional

# Pip
from kw3 import Web3, Busd, Wbnb

# Local
from .reserves_wbnb_busd import ReservesWbnbBusd

from ..core.pancakeswap_liquidity_pool import PancakeswapLiquidityPool

from ...constants import Constants

# -------------------------------------------------------------------------------------------------------------------------------- #



# -------------------------------------------- class: PancakeswapWbnbBusdLiquidityPool ------------------------------------------- #

class PancakeswapWbnbBusdLiquidityPool(PancakeswapLiquidityPool):

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,
        web3: Web3
    ):
        super().__init__(
            web3=web3,
            address=Constants.ADDRESS_LIQUIDITY_POOL_WBNB_BUSD
        )


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    # Custom

    def wbnbAddress(self) -> Optional[str]:
        return self.token0Address()

    def busdAddress(self) -> Optional[str]:
        return self.token1Address()

    def wbnb(self) -> Optional[Wbnb]:
        return self.token0()

    def busd(self) -> Optional[Busd]:
        return self.token1()

    def getReserves(self) -> Optional[ReservesWbnbBusd]:
        return self._getReserves(ReservesWbnbBusd)

    def priceWbnbCumulativeLast(self) -> int:
        return self.price0CumulativeLast()

    def priceBusdCumulativeLast(self) -> int:
        return self.price1CumulativeLast()

    def wbnbPrice(self) -> Optional[float]:
        '''BUSD for 1 BNB'''
        return self.token0Price()

    def busdPrice(self) -> Optional[float]:
        '''BNB for 1 BUSD'''
        return self.token1Price()

    def toWbnb(
        self,
        busdAmount: int
    ) -> Optional[float]:
        '''BNB for "busdAmount" Busd'''
        return self.toToken0(busdAmount)

    def toBusd(
        self,
        wbnbAmount: int
    ) -> Optional[float]:
        '''Busd for "wbnbAmount" BNB'''
        return self.toToken1(wbnbAmount)


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional

# Pip
from kw3 import Web3, Bep20, Busd
from kw3.constants import Constants

# Local
from .reserves_busd import ReservesBusd

from ..core.pancakeswap_liquidity_pool import PancakeswapLiquidityPool

# -------------------------------------------------------------------------------------------------------------------------------- #



# ---------------------------------------------- class: PancakeswapBusdLiquidityPool --------------------------------------------- #

class PancakeswapBusdLiquidityPool(PancakeswapLiquidityPool):

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,
        web3: Web3,
        address: str
    ):
        super().__init__(
            web3=web3,
            address=address
        )


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    # Custom

    def tokenAddress(self) -> Optional[str]:
        return self.token().address

    def token(self) -> Optional[Bep20]:
        return self.quoteToken() if self._is_base_busd() else self.baseToken()

    def busdAddress(self) -> Optional[str]:
        return self.busd().address

    def busd(self) -> Optional[Busd]:
        return self.baseToken() if self._is_base_busd() else self.quoteToken()

    def priceTokenCumulativeLast(self) -> int:
        return self.quoteTokenPriceCumulativeLast() if self._is_base_busd() else self.baseTokenPriceCumulativeLast()

    def priceBusdCumulativeLast(self) -> int:
        return self.baseTokenPriceCumulativeLast() if self._is_base_busd() else self.quoteTokenPriceCumulativeLast()

    def tokenPrice(self) -> Optional[float]:
        '''BUSD for 1 Token'''
        return self.quoteTokenPrice() if self._is_base_busd() else self.baseTokenPrice()

    def busdPrice(self) -> Optional[float]:
        '''Token for 1 BUSD'''
        return self.baseTokenPrice() if self._is_base_busd() else self.quoteTokenPrice()

    def toToken(
        self,
        busdAmount: int
    ) -> Optional[float]:
        '''Token for "busdAmount" BUSD'''
        return (self.toQuoteToken if self._is_base_busd() else self.toBaseToken)(busdAmount)

    def toBusd(
        self,
        tokenAmount: int
    ) -> Optional[float]:
        '''BUSD for "tokenAmount" Token'''
        return (self.toBaseToken if self._is_base_busd() else self.toQuoteToken)(tokenAmount)


    def getReserves(self) -> Optional[ReservesBusd]:
        return self._getReserves(
            ReservesBusd,
            is_base_busd=self._is_base_busd()
        )

    def tokenHoldings(self) -> Optional[int]:
        reserves = self.getReserves()

        return reserves.reserveToken if reserves else None

    def busdHoldings(self) -> Optional[int]:
        reserves = self.getReserves()

        return reserves.reserveBusd if reserves else None


    # ---------------------------------------------------- Private methods --------------------------------------------------- #

    def _is_base_busd(self) -> bool:
        if not hasattr(self, '__is_base_busd'):
            self.__is_base_busd = self.token0().address == Constants.BUSD.ADDRESS

        return self.__is_base_busd


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional

# Pip
from kw3 import WrappedContract, Web3
from kw3.constants import Constants as KW3Constants

# Local
from ._abi import pancakeswap_factory_abi

from ...liquidity_pool import PancakeswapLiquidityPool, PancakeswapBusdLiquidityPool, PancakeswapWbnbLiquidityPool
from ...constants import Constants

# -------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------- class: PancakeswapFactory -------------------------------------------------- #

class PancakeswapFactory(WrappedContract):

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,
        web3: Web3
    ):
        super().__init__(
            web3=web3,
            address=Constants.ADDRESS_PANCAKESWAP_FACTORY,
            abi=pancakeswap_factory_abi
        )


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    # Forwarders

    def liquidityPoolAddressesLength(self) -> int:
        return self.functions.allPairsLength().call()

    def liquidityPoolAddressAtIndex(
        self,
        index: int
    ) -> str:
        return self.functions.allPairs(index).call()

    def liquidityPoolAtIndex(
        self,
        index: int
    ) -> PancakeswapLiquidityPool:
        return PancakeswapBusdLiquidityPool(
            web3=self._web3,
            address=self.liquidityPoolAddressAtIndex(
                index=index
            )
        )


    # Custom

    def getPairAddress(
        self,
        address0: str,
        address1: str
    ) -> Optional[str]:
        return self.functions.getPair(
            Web3.toChecksumAddress(address0),
            Web3.toChecksumAddress(address1)
        ).call()

    def getPair(
        self,
        address0: str,
        address1: str
    ) -> Optional[PancakeswapLiquidityPool]:
        return self.__getPair(
            PancakeswapLiquidityPool,
            address0=address0,
            address1=address1
        )

    def getWbnbPair(
        self,
        token_address: str
    ) -> Optional[PancakeswapWbnbLiquidityPool]:
        return self.__getPair(
            PancakeswapWbnbLiquidityPool,
            address0=KW3Constants.WBNB.ADDRESS,
            address1=token_address
        )

    def getBusdPair(
        self,
        token_address: str
    ) -> Optional[PancakeswapBusdLiquidityPool]:
        return self.__getPair(
            PancakeswapBusdLiquidityPool,
            address0=KW3Constants.BUSD.ADDRESS,
            address1=token_address
        )


    # ---------------------------------------------------- Private methods --------------------------------------------------- #

    def __getPair(
        self,
        _type,
        address0: str,
        address1: str
    ) -> Optional[PancakeswapLiquidityPool]:
        pair_address = self.getPairAddress(address0, address1)

        return _type(
            self._web3,
            pair_address
        ) if pair_address else None


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional, Union, Tuple

# Pip
from kw3 import Web3
from kw3 import WrappedContract, Bep20

# Local
from ._abi import pancakeswap_liquidity_pool_abi
from .reserves import Reserves

from ..busd import PancakeswapBusdLiquidityPool
from ..wbnb import PancakeswapWbnbLiquidityPool

# -------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------ class: PancakeswapLiquidityPool ----------------------------------------------- #

class PancakeswapLiquidityPool(Bep20):

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,
        web3: Web3,
        address: str
    ):
        super().__init__(
            web3=web3,
            address=address,
            abi=pancakeswap_liquidity_pool_abi
        )

        self.__token0_address = None
        self.__token0 = None
        self.__token1_address = None
        self.__token1 = None


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    # "Casters"

    def toWbnbLiquidityPool(self) -> PancakeswapWbnbLiquidityPool:
        return self.__castedSelf(PancakeswapWbnbLiquidityPool)

    def toBusdLiquidityPool(self) -> PancakeswapBusdLiquidityPool:
        return self.__castedSelf(PancakeswapBusdLiquidityPool)

    
    # Custom

    def token0Address(self) -> Optional[str]:
        if not self.__token0_address:
            self.__token0_address = self.functions.token0().call()

        return self.__token0_address

    # alias
    baseTokenAddreess = token0Address


    def token0(self) -> Optional[Bep20]:
        if not self.__token0:
            token0Address = self.token0Address()

            self.__token0 = Bep20(self._web3, token0Address)

        return self.__token0

    # alias
    baseToken = token0


    def token1Address(self) -> Optional[str]:
        if not self.__token1_address:
            self.__token1_address = self.functions.token1().call()

        return self.__token1_address

    # alias
    quoteTokenAddreess = token1Address


    def token1(self) -> Optional[Bep20]:
        if not self.__token1:
            token1Address = self.token1Address()

            self.__token1 = Bep20(self._web3, token1Address)

        return self.__token1

    # alias
    quoteToken = token1


    def token0Price(self) -> Optional[float]:
        '''token1 for 1 token0'''
        reserves = self.getReserves()

        return reserves.token0Price(
            token_0_decimals=self.token0().decimals(),
            token_1_decimals=self.token1().decimals(),
        ) if reserves else None

    # alias
    baseTokenPrice = token0Price


    def token1Price(self) -> Optional[float]:
        '''token0 for 1 token1'''
        reserves = self.getReserves()

        return reserves.token1Price(
            token_0_decimals=self.token0().decimals(),
            token_1_decimals=self.token1().decimals(),
        ) if reserves else None

    # alias
    quoteTokenPrice = token1Price


    def toToken0(
        self,
        token1Amount: int
    ) -> Optional[float]:
        '''How much token0 for "token1Amount" token1'''
        token1Price = self.token1Price()

        return token1Amount * token1Price if token1Price else None

    # alias
    toBaseToken = toToken0


    def toToken1(
        self,
        token1Amount: int
    ) -> Optional[float]:
        '''How much token1 for "token0Amount" token0'''
        token0Price = self.token0Price()

        return token1Amount * token0Price if token0Price else None

    # alias
    toQuoteToken = toToken1


    # Forwarders

    def price0CumulativeLast(self) -> int:
        return self.functions.price0CumulativeLast().call()

    # alias
    baseTokenPriceCumulativeLast = price0CumulativeLast


    def price1CumulativeLast(self) -> int:
        return self.functions.price1CumulativeLast().call()

    # alias
    quoteTokenPriceCumulativeLast = price1CumulativeLast


    def getReserves(self) -> Optional[Reserves]:
        return self._getReserves(Reserves)


    # ---------------------------------------------------- Private methods --------------------------------------------------- #

    def _getReserves(
        self,
        _type,
        **extra_args
    ) -> Optional[Reserves]:
        res = self.functions.getReserves().call()

        return _type(
            reserve0=res[0],
            reserve1=res[1],
            blockTimestampLast=res[2],
            **extra_args
        ) if res else None

    def __castedSelf(
        self,
        _type
    ):
        return _type(
            web3=self._web3,
            address=self.address
        )


# -------------------------------------------------------------------------------------------------------------------------------- #